/*
 * generated by Xtext
 */
package fr.inria.phoenix.diasuite.generator

import com.google.inject.Inject
import fr.inria.phoenix.diasuite.diaswarm.AbstractElement
import fr.inria.phoenix.diasuite.diaswarm.ContextDef
import fr.inria.phoenix.diasuite.diaswarm.ControllerDef
import fr.inria.phoenix.diasuite.diaswarm.DeviceType
import fr.inria.phoenix.diasuite.diaswarm.DomainModel
import fr.inria.phoenix.diasuite.diaswarm.EnumDef
import fr.inria.phoenix.diasuite.diaswarm.InteractionContractDef
import fr.inria.phoenix.diasuite.diaswarm.PushInteractionContract
import fr.inria.phoenix.diasuite.diaswarm.StructDef
import fr.inria.phoenix.diasuite.generator.template.IConstant
import fr.inria.phoenix.diasuite.generator.template.context.DiaSwarmContext
import fr.inria.phoenix.diasuite.generator.template.controller.DiaSwarmController
import fr.inria.phoenix.diasuite.generator.template.datatype.DiaSwarmDatatype
import fr.inria.phoenix.diasuite.generator.template.deployment.DiaSwarmDeployment
import fr.inria.phoenix.diasuite.generator.template.device.DiaSwarmDevice
import java.util.HashMap
import java.util.Map
import org.eclipse.emf.ecore.resource.Resource
import org.eclipse.xtext.generator.IFileSystemAccess
import org.eclipse.xtext.generator.IGenerator

import static extension fr.inria.phoenix.diasuite.generator.template.context.DiaSwarmContextHelper.*
import static extension fr.inria.phoenix.diasuite.generator.template.controller.DiaSwarmControllerHelper.*
import static extension fr.inria.phoenix.diasuite.generator.template.datatype.DiaSwarmDatatypeHelper.*
import static extension fr.inria.phoenix.diasuite.generator.template.deployment.DiaSwarmDeploymentHelper.*
import static extension fr.inria.phoenix.diasuite.generator.template.device.DiaSwarmDeviceHelper.*

/**
 * Generates code from your model files on save.
 * 
 * see http://www.eclipse.org/Xtext/documentation.html#TutorialCodeGeneration
 */
class DiaSwarmGenerator implements IGenerator {

	@Inject extension DiaSwarmDatatype
	@Inject extension DiaSwarmDevice
	@Inject extension DiaSwarmContext
	@Inject extension DiaSwarmController
	@Inject extension DiaSwarmDeployment

	override void doGenerate(Resource resource, IFileSystemAccess fsa) {
		// Here we begin to generate code for DiaSwarm components and data types.
		for (entity : resource.allContents.toIterable.filter(typeof(AbstractElement)))
			entity.generateCode(fsa)

		// Here we begin to generate deployment code for a DiaSwarm application.
		for (entity : resource.allContents.toIterable.filter(typeof(DomainModel)))
			entity.generateCode(fsa)
	}

	/**
	 * Generates code for DiaSwarm structures.
	 */
	def dispatch generateCode(StructDef structure, IFileSystemAccess fsa) {
		fsa.generateFile(structure.generateFileName, structure.generateStructure)
	}

	/**
	 * Generates code for DiaSwarm enumerations.
	 */
	def dispatch generateCode(EnumDef enumeration, IFileSystemAccess fsa) {
		fsa.generateFile(enumeration.generateFileName, enumeration.generateEnumeration)
	}

	/**
	 * Generates code for DiaSwarm devices.
	 */
	def dispatch generateCode(DeviceType device, IFileSystemAccess fsa) {
		fsa.generateFile(device.generateFileName, device.generateDevice)

		// Here we generate individual classes for every device source
		device.allSources.forEach([ source |
			if (!source.indices.empty)
				fsa.generateFile(source.generateFileName(device, "Indices"), source.generateDeviceSourceIndices(device))
			fsa.generateFile(source.generateFileName(device, "From" + device.name),
				source.generateDeviceSourceValue(device))
		])
	}

	/**
	 * Generates code for DiaSwarm contexts.
	 */
	def dispatch generateCode(ContextDef context, IFileSystemAccess fsa) {
		fsa.generateFile(context.generateFileName, context.generateContext)

		if (!context.getIndices().isEmpty()) {
			fsa.generateFile(context.generateFileNameForContextFeature("Indices"), context.contextIndices)
		}
		fsa.generateFile(context.generateFileNameForContextFeature("Value"), context.contextValue)

		val mapReduceBehavior = context.behaviors.filter [
			(interaction.equals(IConstant.PERIODIC_INTERACTION)) &&
				(contract as PushInteractionContract).providedData.map &&
				(contract as PushInteractionContract).providedData.reduce
		];

		if (!mapReduceBehavior.isEmpty) {
			fsa.generateFile(context.generateFileNameForContextFeature("Job"),
				context.generateJob(mapReduceBehavior.last.contract as PushInteractionContract))
			fsa.generateFile(context.generateFileNameForContextFeature("Adapter"),
				context.generateAdapter(mapReduceBehavior.last.contract as PushInteractionContract))
			if ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.mapType.type != null) {
				// if it is an enumeration
				if ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.mapType.
					type instanceof EnumDef) {
						val enumeration = ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.
							mapType.type as EnumDef);
							fsa.generateFile(enumeration.generateFileName("Writable"),
								enumeration.generateWritableEnumeration)
				}

				// it is a source
				if ((mapReduceBehavior.last.contract as PushInteractionContract).providedData.mapType.
					type instanceof StructDef) {
						val structure = ((mapReduceBehavior.last.contract as PushInteractionContract).
							providedData.mapType.type as StructDef)
						fsa.generateFile(structure.generateFileName("Writable"), structure.generateWritableStructure)
				}
			}
		}
	}

	/**
	 * Generates code for DiaSwarm controllers.
	 */
	def dispatch generateCode(ControllerDef controller, IFileSystemAccess fsa) {
		fsa.generateFile(controller.generateFileName, controller.generateController)
	}

	/**
	 * Generates application deployment code.
	 */
	def dispatch generateCode(DomainModel model, IFileSystemAccess fsa) {
		var deviceSourceBufferSize = new HashMap<String, Integer>();
		var deviceSourceToGroupingAttributes = new HashMap<String, Map<String, String>>();
		var deviceSourceType = new HashMap<String, String>();

		for (ContextDef context : model.contexts) {
			for (InteractionContractDef behavior : context.behaviors) {
				if (behavior.interaction.equals(IConstant.PERIODIC_INTERACTION)) {
					var requirement = (behavior.contract as PushInteractionContract).providedData
					var sourceRef = requirement.dataSource.sourceRef

					if (sourceRef != null) {
						var deviceName = sourceRef.device.name
						var sourceName = sourceRef.source.name
						var String sourceType = sourceRef.source.dataTypeRef.fullName

						deviceSourceType.put(sourceName + "BufferFor" + deviceName, sourceType)

						println("sourceName : " + sourceName);
						var devicePublishingPeriodicity = sourceRef.periodicity
						var groupingPeriodicity = requirement.periodicity
						var bufferSize = 1

						if (groupingPeriodicity != null)
							bufferSize = groupingPeriodicity.periodicityInSeconds /
								devicePublishingPeriodicity.periodicityInSeconds

						var lastBufferSize = deviceSourceBufferSize.get(sourceName + "BufferFor" + deviceName)

						if (lastBufferSize == null || lastBufferSize < bufferSize)
							deviceSourceBufferSize.put(sourceName + "BufferFor" + deviceName, bufferSize)

						var groupingAttribute = requirement.groupingAttribute

						if (groupingAttribute != null) {
							var groupingAttributeName = groupingAttribute.name;
							var String groupingAttributeType;
							var attributes = deviceSourceToGroupingAttributes.get(sourceName + "BufferFor" + deviceName)

							if (attributes == null) {
								attributes = new HashMap<String, String>();
								deviceSourceToGroupingAttributes.put(sourceName + "BufferFor" + deviceName, attributes)
							}

							if (groupingAttribute.attrTypeRef.type != null)
								groupingAttributeType = groupingAttribute.attrTypeRef.fullName
							else
								groupingAttributeType = groupingAttribute.attrTypeRef.primitiveTypeRef.literal

							attributes.put(groupingAttributeName, groupingAttributeType)
						}
					}
				}
			}
		}

		fsa.generateFile(model.generateDeploymentFileName, model.generateDeploymentCode(deviceSourceBufferSize.keySet))

		for (bufferName : deviceSourceBufferSize.keySet) {
			var bufferSize = deviceSourceBufferSize.get(bufferName)
			var attributes = deviceSourceToGroupingAttributes.get(bufferName)
			var sourceType = deviceSourceType.get(bufferName)

			println("Buffer name : " + bufferName)
			println("Buffer size : " + bufferSize)

			println("---------------------------------------------")

			fsa.generateFile(model.generateBufferFileName(bufferName),
				model.generateBufferForDeviceSource(bufferName, (bufferName.split("BufferFor")).get(0), sourceType,
					bufferSize, attributes))
		}
	}

}