package fr.inria.phoenix.diasuite.generator.template

import fr.inria.phoenix.diasuite.diaswarm.ContextDef
import fr.inria.phoenix.diasuite.diaswarm.ControllerBehaviorDef
import fr.inria.phoenix.diasuite.diaswarm.ControllerDef
import fr.inria.phoenix.diasuite.diaswarm.DataTypeRef
import fr.inria.phoenix.diasuite.diaswarm.DeviceType
import fr.inria.phoenix.diasuite.diaswarm.DomainModel
import fr.inria.phoenix.diasuite.diaswarm.PrimitiveTypeRef
import fr.inria.phoenix.diasuite.diaswarm.SourceDef
import fr.inria.phoenix.diasuite.diaswarm.SourceType
import fr.inria.phoenix.diasuite.diaswarm.StructDef
import fr.inria.phoenix.diasuite.diaswarm.VariableDef
import java.util.HashSet
import java.util.List
import java.util.Map
import java.util.Set
import org.eclipse.emf.ecore.EObject

import static extension fr.inria.phoenix.diasuite.generator.template.datatype.DiaSwarmDatatypeHelper.*

class Utils {

	def static <T> transformToIndexedSet(Iterable<T> i) {
		val Map<Integer, T> map = <Integer, T>newHashMap()
		var counter = 0

		for (T e : i) {
			map.put(counter, e)
			counter = counter + 1
		}

		val Set<Map.Entry<Integer, T>> res = map.entrySet
		res
	}

	/* Returns the class name for a context */
	def static String classNameSpace(ContextDef context) {
		return "/Context/" + context.name + "/";
	}

	def static String plural(String name) {
		if (name.endsWith("y") && !name.endsWith("way")) {
			return name.substring(0, name.length() - 1) + "ies";
		}
		return name + "s";
	}

	def static String parametersInOrder(Iterable<VariableDef> variables, String argName) {
		index(variables).map(['''(«fullname(value.dataTypeRef)») «argName»[«key»]''']).join(",\n        ");
	}

	def static <T> index(Iterable<T> i) {
		val Map<Integer, T> map = <Integer, T>newHashMap()
		var counter = 0
		for (T e : i) {
			map.put(counter, e)
			counter = counter + 1
		}
		val Set<Map.Entry<Integer, T>> res = map.entrySet
		res
	}

	def static String fullname(DataTypeRef dataType) {
		return if (dataType.list) {
			'''java.util.List<«dataType.typeName»>'''
		} else {
			dataType.typeName
		}
	}

	def static ControllerDef getControllerContainer(EObject elt) {
		return ( getParentOfClass(elt, ControllerDef) as ControllerDef);
	}

	def static ControllerBehaviorDef getBehaviorActionListContainer(EObject elt) {
		return ( getParentOfClass(elt, ControllerBehaviorDef) as ControllerBehaviorDef);
	}

	def static EObject getParentOfClass(EObject elt, Class<? extends EObject> clazz) {
		if (elt instanceof DomainModel) {
			return null;
		}
		if (clazz.isInstance(elt)) {
			return elt;
		} else {
			return getParentOfClass(elt.eContainer(), clazz);
		}
	}

//	/**
//	 * Return the full list of action trigger (recursive) of a device
//	 * 
//	 * @param device
//	 *            The device to extract actions from
//	 * @return A map of order => action containing couples of orders and their containing action
//	 */
//	def static Map<OrderDef, ActionImpl> getAllActions(DeviceType device) {
//		println("SUPERDEVICES : " + device.allSuperDevices.toString);
//		return null;
//	}
//	def static Map<OrderDef, ActionImpl> getAllActions(DeviceType device) {
//		var Set<String> processed = new HashSet<String>();
//		var Map<OrderDef, ActionImpl> orders = new HashMap<OrderDef, ActionImpl>();
//		for (ActionImpl ai : device.getActions()) {
//			for (OrderDef od : ai.getAction().getOrders()) {
//				processed.add(od.getName());
//				orders.put(od, ai);
//			}
//		}
//		for (DeviceType dev : getAllSuperDevices(device)) {
//			for (ActionImpl ai : dev.getActions()) {
//				for (OrderDef od : ai.getAction().getOrders()) {
//					if (!processed.contains(od.getName())) {
//						processed.add(od.getName());
//						orders.put(od, ai);
//					}
//				}
//			}
//		}
//		return orders;
//	}
	/**
	 * Return the list of super devices (recursive) of a device
	 * 
	 * @param device
	 *            The device to extract super device from
	 * @return The set of super devices
	 */
	def static Set<DeviceType> getAllSuperDevices(DeviceType device) {
		return getAllSuperDevices(device, new HashSet<DeviceType>());
	}

	def static Set<DeviceType> getAllSuperDevices(DeviceType device, Set<DeviceType> result) {
		if (device.superDevice != null)
			if (!result.contains(device.superDevice)) { // cycle avoidance
				result.add(device.superDevice);
				if (device.superDevice != null)
					result.addAll(getAllSuperDevices(device.superDevice, result));
			}

		return result;
	}

	def static StructDef getStructDefContainer(EObject elt) {
		return ( getParentOfClass(elt, StructDef) as StructDef);
	}

	def static ContextDef getContextContainer(EObject elt) {
		return ( getParentOfClass(elt, ContextDef) as ContextDef);
	}

	def static DomainModel getDocumentContainer(EObject elt) {
		if (elt instanceof DomainModel) {
			return ( elt as DomainModel);
		} else {
			return getDocumentContainer(elt.eContainer());
		}
	}

	/**
	 * Returns true if src1 and src2 does not have the same name or if src1 and src2 have exactly the same signature
	 * 
	 * @param src1
	 *            The first source definition to test
	 * @param src2
	 *            the second source definition to test
	 * @return true if src1 and src2 can co-exists in the same device
	 */
	def static boolean isCompatibleSource(SourceDef src1, SourceDef src2) {
		if (src1 == src2)
			return true;
		if (!src1.getName().equals(src2.getName()))
			return true;
		if (!isSameType(src1.getDataTypeRef(), src2.getDataTypeRef()))
			return false;
		if (src1.getIndices().size() != src2.getIndices().size())
			return false;
		return isCompatibleVariableDefList(src1.getIndices(), src2.getIndices());
	}

	/**
	 * Returns true if varDefs1 and varDefs2 have exactly the same signature
	 * 
	 * @param varDefs1
	 *            A list of variable definitions to compare
	 * @param varDefs2
	 *            A second list of variable definitions to compare
	 * @return true if varDefs1 and varDefs2 have the same length, with element of the same names and same types
	 */
	def static boolean isCompatibleVariableDefList(List<VariableDef> varDefs1, List<VariableDef> varDefs2) {
		for (var i = 0; i < varDefs1.size(); i++) {
			var VariableDef varDef1 = varDefs1.get(i);
			var VariableDef varDef2 = varDefs2.get(i);
			if (!varDef1.getName().equals(varDef2.getName()))
				return false;
			if (!isSameType(varDef1.getDataTypeRef(), varDef2.getDataTypeRef()))
				return false;
		}
		return true;
	}

	/**
	 * Return true if type1 and type2 represent the same type
	 * 
	 * @param type1
	 *            The first type to compare
	 * @param type2
	 *            The second type to compare
	 * @return true if type1 represent the same type than type2
	 */
	def static boolean isSameType(DataTypeRef type1, DataTypeRef type2) {
		return extractDatatypeName(type1).equals(extractDatatypeName(type2));
	}

	/**
	 * Convert a DataTypeRef structure describing a data type into its textual representation
	 * 
	 * For instance, if dataTypeRef is a list of integer then the result will be "Integer[]"
	 * 
	 * @param dataTypeRef
	 *            The data type structure to convert
	 * @return A string representation of the data type represented by <b>dataTypeRef</b>
	 */
	def static String extractDatatypeName(DataTypeRef dataTypeRef) {
		var String dataType = "";
		val SourceType sourceType = dataTypeRef.getType();
		if (sourceType != null) {
			dataType = sourceType.getName();
		} else if (dataTypeRef.getPrimitiveTypeRef() != null) {
			dataType = dataTypeRef.getPrimitiveTypeRef().getLiteral();
		}
		if (dataType == null) {
			dataType = "";
		}
		// is List ?
		if (dataTypeRef.isList()) {
			dataType += "[]";
		}
		return dataType;
	}

	/**
	 * Returns a string representing the data-type of the field 
	 * provided by parameter.
	 */
//	def static getHadoopTypeName(PrimitiveTypeRef dataType) {
//		switch dataType {
//			case PrimitiveTypeRef::BOOLEAN: IConstant.BOOLEAN_WRITABLE
//			case PrimitiveTypeRef::INTEGER: IConstant.INTEGER_WRITABLE
//			case PrimitiveTypeRef::FLOAT: IConstant.FLOAT_WRITABLE
//			case PrimitiveTypeRef::STRING: IConstant.TEXT
//			case PrimitiveTypeRef::BINARY: IConstant.BYTE_WRITABLE
//		}
//	}
	def static getHadoopTypeName(DataTypeRef dataType) {
		return if (dataType.type != null) {
			dataType.getTypeName + "Writable";
		} else {
			switch dataType.primitiveTypeRef {
				case PrimitiveTypeRef::BOOLEAN: IConstant.BOOLEAN_WRITABLE
				case PrimitiveTypeRef::INTEGER: IConstant.INTEGER_WRITABLE
				case PrimitiveTypeRef::FLOAT: IConstant.FLOAT_WRITABLE
				case PrimitiveTypeRef::STRING: IConstant.TEXT
				case PrimitiveTypeRef::BINARY: IConstant.BYTE_WRITABLE
			}
		}
	}
}