package fr.inria.phoenix.diasuite.generator.template.controller

import fr.inria.phoenix.diasuite.diaswarm.ActionRef
import fr.inria.phoenix.diasuite.diaswarm.ContextDef
import fr.inria.phoenix.diasuite.diaswarm.ControllerBehaviorDef
import fr.inria.phoenix.diasuite.diaswarm.ControllerDef
import fr.inria.phoenix.diasuite.generator.template.DiaSwarmEntity
import fr.inria.phoenix.diasuite.generator.template.IConstant

import static extension fr.inria.phoenix.diasuite.generator.template.GenericTemplate.*
import static extension fr.inria.phoenix.diasuite.generator.template.Utils.*
import static extension fr.inria.phoenix.diasuite.generator.template.context.DiaSwarmContextHelper.*
import static extension fr.inria.phoenix.diasuite.generator.template.controller.DiaSwarmControllerHelper.*
import static extension fr.inria.phoenix.diasuite.generator.template.device.DiaSwarmDeviceHelper.*

class DiaSwarmController extends DiaSwarmEntity {

	def generateController(ControllerDef controller) {
		controller.abstractClassForController()
	}
	
//    override visitControllerDef(ControllerDef controller) {
//        val packageName = controller.packageName
//        generateJavaFile("Abstract" + controller.name, packageName,
//                controller.abstractClass);
//    }
    
    def imports(ControllerDef controller) '''
        import fr.inria.diagen.core.ServiceConfiguration;
        «IF controller.behaviors.exists[actions.exists[!device.allAttributes.nullOrEmpty]] »
        import fr.inria.diagen.core.service.composite.CompositeException;
        «ENDIF»
        import fr.inria.diagen.core.exception.InvocationException;
        import fr.inria.diagen.core.network.RemoteServiceInfo;
        import fr.inria.diagen.core.service.local.Service;
        import fr.inria.diagen.core.service.proxy.Proxy;
        import fr.inria.diagen.core.network.DataDeliveryEnum;
        import fr.inria.diagen.core.network.DataDeliveryModel;
        import fr.inria.diagen.core.service.local.ServiceProperties;
        
        import java.util.Map;
        import java.util.HashMap;
        
        «FOR context : controller.allContexts.sortWith([n1, n2 | n1.name.compareTo(n2.name)])»
        import «context.packageName».«context.name»Value;
        «ENDFOR»
    '''

    def abstractClassForController(ControllerDef controller) ''' 
        package «controller.packageName»;
        
        «controller.imports»
        
«««        «controller.javadoc»
        @SuppressWarnings("all")
        public abstract class Abstract«controller.name.toFirstUpper» extends Service {
        	
        	«controller.classAttributes»
            
            public Abstract«controller.name.toFirstUpper»(ServiceConfiguration serviceConfiguration) {
                super("/Controller/«controller.name.toFirstUpper»/", serviceConfiguration);
                addToOutProperties(ServiceProperties.DELIVERY.toString(),
    				deliveryModels);
            }
            
            «controller.serviceClassMethods»
            
            «controller.interactionContracts»
            «FOR behaviorDef : controller.behaviors»
                
                «behaviorDef.discoverCodeForBehavior»
            «ENDFOR»
        }
    '''
    
    def classAttributes(ControllerDef controller){ 
    	'''
    	private static final Map<String, Object> deliveryModels;
    		
    		static {
    			deliveryModels = new HashMap<String, Object>();
    			«FOR behavior : controller.behaviors»
    					«getDeliveryModelsMap
    						(IConstant.EVENT_DRIVEN_INTERACTION, 
    							behavior.activationCondition.name,
    								0, null, 0)»
    			«ENDFOR»
    		}
    	'''
    }
    
    def serviceClassMethods(ControllerDef controller) '''
        // Methods from the Service class
        @Override
        protected final void internalPostInitialize() {
            «FOR v : controller.providedContexts.sortWith([n1, n2 | n1.name.compareTo(n2.name)])»
                subscribeValue("«v.name»", "«v.classNameSpace»"); // subscribe to «v.name» context
            «ENDFOR»
            postInitialize();
        }
        
        @Override
        public final void valueReceived(java.util.Map<String, Object> properties, RemoteServiceInfo source, String eventName, Object value, Object... indexes) {
            «FOR v : controller.behaviors.map[activationCondition].sortBy[n | n.name]»
                «v.valueReceivedFromContext»
            «ENDFOR»
        }
        // End of methods from the Service class
    '''
    
    def valueReceivedFromContext(ContextDef contextRef) '''
        if (eventName.equals("«contextRef.name»") && source.isCompatible("«contextRef.classNameSpace»")) {
            «contextRef.name»Value «contextRef.name.toFirstLower»Value = new «contextRef.name»Value((«contextRef.dataTypeRef.fullname») value«
                if(contextRef.indices.nullOrEmpty) ""
                else ",\n        " + contextRef.indices.parametersInOrder("indexes")
            »);
            
            on«contextRef.name»(«contextRef.name.toFirstLower»Value, new Discover());
        }
    '''
    
    def interactionContracts(ControllerDef controller) '''
        // Interaction contract implementation
        «FOR v : controller.behaviors»
            /**
             * This method is called when the <code>«v.activationCondition.name»</code> context publishes a value.
             * 
             * <pre>
«««             * «v.stripComments.replace("\n", "\n* ")»
             * </pre>
             * 
             * @param «v.activationCondition.name.toFirstLower» the value of the <code>«v.activationCondition.name»</code> context.
             * @param discover a discover object to get context values and action methods
             */
            protected abstract void on«v.activationCondition.name»(«v.activationCondition.name»Value «v.activationCondition.name.toFirstLower», Discover discover);

        «ENDFOR»
        // End of interaction contract implementation
    '''


    
    /* Create the discoverer class for one behavior */
    def discoverCodeForBehavior(ControllerBehaviorDef behaviorDef) {
        return behaviorDef.controllerContainer.discoverCodeForBehavior(behaviorDef.activationCondition.name,
                           /*behaviorDef.stripComments*/"",
                           behaviorDef.dataRequirements,
                           behaviorDef.behaviorActionListContainer.actions)
    }
    
    /* Create the discoverer class for one behavior */
    def discoverCodeForBehavior(ControllerDef controller, String name, String comment, Iterable<ContextDef> contexts, Iterable<ActionRef> actions) {
        val devices = actions.map[device].toSet; 
        return '''
        // Discover object for «name»
        «discovererClassForBehavior(name, "Abstract" + controller.name, comment, devices, contexts)»
        «FOR dev : devices»
            
            «dev.discovererClassForDevice("For" + name, '''
            Discover object that will exposes the <code>«dev.name»</code> devices to execute action on for the
            <code>when provided «name»</code> interaction contract.''')»
            
            «dev.compositeClassForDevice("For" + name, null, actions.filter[device == dev].map[action], '''
            A composite of several <code>«dev.name»</code> devices to execute action on for the
            <code>when provided «name»</code> interaction contract.''')»
            
            «dev.proxyClassForDevice("For" + name, null, null, actions.filter[device == dev].map[action], '''
            A proxy to one <code>«dev.name»</code> device to execute action on for the
            <code>when provided «name»</code> interaction contract.''')»
        «ENDFOR»
        // End of discover object for «name»
        '''
    }
}
