package fr.inria.phoenix.diasuite.generator.template.datatype

import fr.inria.phoenix.diasuite.diaswarm.EnumDef
import fr.inria.phoenix.diasuite.diaswarm.PrimitiveTypeRef
import fr.inria.phoenix.diasuite.diaswarm.StructDef
import fr.inria.phoenix.diasuite.generator.template.DiaSwarmEntity

import static extension fr.inria.phoenix.diasuite.generator.template.datatype.DiaSwarmDatatypeHelper.*
import static extension fr.inria.phoenix.diasuite.generator.template.Utils.*

class DiaSwarmDatatype extends DiaSwarmEntity {

	def generateEnumeration(EnumDef enumDef) '''
		package «enumDef.packageName»;        
		
		public enum «enumDef.name.toFirstUpper» {
		    «enumDef.constants.join(", ")»
		}
	'''

	def generateWritableEnumeration(EnumDef enumDef) '''
		package «enumDef.packageName»;        
		
		public enum «enumDef.name.toFirstUpper» {
		    «enumDef.constants.join(", ")»
		}
	'''

	/**
	 * Generates the java code for the structure provided by parameter. 
	 */
	def generateWritableStructure(StructDef structDef) '''
		package «structDef.packageName»;
		
		import java.io.DataInput;
		import java.io.DataOutput;
		import java.io.IOException;
		
		«FOR field : structDef.structFields»
		import «field.dataTypeRef.hadoopTypeName»;
		«ENDFOR»
		import org.apache.hadoop.io.Writable;
		
		public class «structDef.name.toFirstUpper»Writable implements Writable {
			
			«FOR field : structDef.structFields»
				«field.dataTypeRef.hadoopTypeName» «field.name»;
			«ENDFOR»
			
			«val constructorEmptyWritableParams = structDef.structFields.map["new " + dataTypeRef.getHadoopTypeName().split('\\.').last + "()" ].join(",\n")»
			«val constructorDefinedWritableParams = structDef.structFields.map["new " + dataTypeRef.getHadoopTypeName().split('\\.').last + "(" + name + ")" ].join(",\n")»
			«val constructorDefinedTypeParam = structDef.structFields.map["new " + dataTypeRef.getHadoopTypeName().split('\\.').last + "(" + structDef.name.toFirstLower + "." + "get" + name.toFirstUpper + "()" + ")" ].join(",\n")»
			«val constructorParams = structDef.structFields.map[dataTypeRef.getTypeName().split('\\.').last + " " + name ].join(",\n")»
			«val constructorWritableParams = structDef.structFields.map[dataTypeRef.getHadoopTypeName().split('\\.').last + " " + name ].join(",\n")»
			«val writableFieldsToJava = structDef.structFields.map[if(dataTypeRef.getHadoopTypeName().split('\\.').last.equals("Text")) (name + ".toString()") else (name + ".get()") ].join(",\n")»
			
			public «structDef.name.toFirstUpper»Writable() {
				set(«constructorEmptyWritableParams»);
			}
		
			public «structDef.name.toFirstUpper»Writable(«constructorParams») {
				set(«constructorDefinedWritableParams»);
			}
			
			public «structDef.name.toFirstUpper»Writable(«structDef.name» «structDef.name.toFirstLower») {
				set(«constructorDefinedTypeParam»);
			}
		
			private void set(«constructorWritableParams») {
				«FOR field : structDef.structFields»
				this.«field.name» = «field.name»;
				«ENDFOR»
			}
			
			public «structDef.name.toFirstUpper» get() {
				return new «structDef.name.toFirstUpper»(«writableFieldsToJava»);
			}
		
			«FOR field : structDef.structFields»
			public «field.dataTypeRef.hadoopTypeName» get«field.name.toFirstUpper»() {
				return this.«field.name»;
			}
			«ENDFOR»

			@Override
			public void readFields(DataInput in) throws IOException {
				«FOR field : structDef.structFields»
				«field.name».readFields(in);
				«ENDFOR»
			}
		
			@Override
			public void write(DataOutput out) throws IOException {
				«FOR field : structDef.structFields»
				«field.name».write(out);
				«ENDFOR»
			}
		}
	'''

	/**
	 * Generates the java code for the structure provided by parameter. 
	 */
	def generateStructure(StructDef structDef) '''
		package «structDef.packageName»;
		
		import java.io.Serializable;        
		
		public class «structDef.name.toFirstUpper» implements Serializable {
		    
		    private static final long serialVersionUID = 0;
		    
		    public «structDef.name.toFirstUpper»() {
		    }
		
		    «IF ! structDef.structFields.isEmpty()»
			public «structDef.name.toFirstUpper»(«structDef.structFields.map[dataTypeRef.fullName + " " + name.toFirstLower].join(",\n        ")») {
			    «FOR field: structDef.structFields»
			    	this.«field.name.toFirstLower» = «field.name.toFirstLower»;
			    «ENDFOR»
			}
		    «ENDIF»
		    «FOR field : structDef.structFields»
			
			/********** Generated code for field «field.name» **********/
			
			private «field.dataTypeRef.fullName» «field.name.toFirstLower»;
			
			public «field.dataTypeRef.fullName» get«field.name.toFirstUpper»() {
				return «field.name.toFirstLower»;
			}
			
			public void set«field.name.toFirstUpper»(«field.dataTypeRef.fullName» «field.name.toFirstLower») {
				this.«field.name.toFirstLower» = «field.name.toFirstLower»;
			}
			
			/**************************************************/
			
			«ENDFOR» 
			«equalsAndHashCode(structDef.name, structDef.structFields.map[name])»
		
		   @Override
		   public String toString() {
		       return "«structDef.name» [" + 
		           «FOR field : structDef.structFields SEPARATOR "\", \" + "»
		           	«IF field.dataTypeRef.primitiveTypeRef.equals(PrimitiveTypeRef.BINARY)»
		           		"«field.name.toFirstLower».length=" + «field.name.toFirstLower».length +
		           	«ELSE»
		           		"«field.name.toFirstLower»=" + «field.name.toFirstLower» +
		           	«ENDIF»
		           «ENDFOR»
		       "]";
		   }
		}
	'''

}
