package fr.inria.phoenix.diasuite.generator.template.deployment

import fr.inria.phoenix.diasuite.diaswarm.ContextDef
import fr.inria.phoenix.diasuite.diaswarm.ControllerDef
import fr.inria.phoenix.diasuite.diaswarm.DeviceType
import fr.inria.phoenix.diasuite.diaswarm.DomainModel
import fr.inria.phoenix.diasuite.diaswarm.Periodicity
import fr.inria.phoenix.diasuite.diaswarm.TimeUnit
import fr.inria.phoenix.diasuite.generator.template.IConstant
import java.util.ArrayList
import java.util.HashMap
import java.util.List

class DiaSwarmDeploymentHelper {

	/**============================ HELPER METHODS ==============================**/
	def static generateDeploymentFileName(DomainModel model) {
		model.getPackageName.replaceAll("\\.", "/") + "/" + IConstant.DEPLOYMENT_FILE_NAME +
			IConstant.JAVA_EXTENSION_POSTFIX
	}

	def static generateBufferFileName(DomainModel model, String bufferName) {
		model.getPackageName.replaceAll("\\.", "/") + "/" + bufferName.toFirstUpper + IConstant.JAVA_EXTENSION_POSTFIX
	}

	def static getPackageName(DomainModel model) {
		IConstant.BASE_PACKAGE_NAMESPACE + "." + IConstant.DEPLOYMENT_PACKAGE_NAME
	}

	def static<T> Iterable<T> uniq(Iterable<? extends T> original, (T)=>Object func) {
		val values = new HashMap<Object, T>();
		original.forEach(o|values.put(func.apply(o), o));
		return values.values;
	}

	def static Iterable<ContextDef> getContexts(DomainModel model) {
		var List<ContextDef> contexts = new ArrayList<ContextDef>()

		for (element : model.elements)
			if (element instanceof ContextDef)
				contexts.add(element as ContextDef)

		contexts
	}

	def static Iterable<ControllerDef> getControllers(DomainModel model) {
		var List<ControllerDef> controllers = new ArrayList<ControllerDef>()

		for (element : model.elements)
			if (element instanceof ControllerDef)
				controllers.add(element as ControllerDef)

		controllers
	}

	def static Iterable<DeviceType> getDevices(DomainModel model) {
		var List<DeviceType> devices = new ArrayList<DeviceType>()

		for (element : model.elements)
			if (element instanceof DeviceType)
				devices.add(element as DeviceType)

		devices
	}

	def static Integer getPeriodicityInSeconds(Periodicity periodicity) {
		switch (periodicity.timeUnit) {
			case TimeUnit.S: return periodicity.value
			case TimeUnit.MIN: return periodicity.value * 60
			case TimeUnit.HR: return (periodicity.value * 60) * 60
			default: return null
		}
	}

}