/*
 * generated by Xtext
 */
package fr.inria.phoenix.diasuite.validation

import com.google.inject.Inject
import fr.inria.phoenix.diasuite.diaswarm.ActionDef
import fr.inria.phoenix.diasuite.diaswarm.ActionImpl
import fr.inria.phoenix.diasuite.diaswarm.ActionRef
import fr.inria.phoenix.diasuite.diaswarm.AttributeDef
import fr.inria.phoenix.diasuite.diaswarm.ContextDef
import fr.inria.phoenix.diasuite.diaswarm.ControllerBehaviorDef
import fr.inria.phoenix.diasuite.diaswarm.ControllerDef
import fr.inria.phoenix.diasuite.diaswarm.DataRequirement
import fr.inria.phoenix.diasuite.diaswarm.DeviceType
import fr.inria.phoenix.diasuite.diaswarm.DiaswarmPackage
import fr.inria.phoenix.diasuite.diaswarm.DomainModel
import fr.inria.phoenix.diasuite.diaswarm.IncludeSpec
import fr.inria.phoenix.diasuite.diaswarm.InteractionContractDef
import fr.inria.phoenix.diasuite.diaswarm.OrderDef
import fr.inria.phoenix.diasuite.diaswarm.PushInteractionContract
import fr.inria.phoenix.diasuite.diaswarm.SourceDef
import fr.inria.phoenix.diasuite.diaswarm.SourceType
import fr.inria.phoenix.diasuite.diaswarm.StructDef
import fr.inria.phoenix.diasuite.diaswarm.StructFieldDef
import fr.inria.phoenix.diasuite.generator.template.Utils
import java.util.ArrayList
import java.util.HashMap
import java.util.HashSet
import java.util.LinkedList
import java.util.List
import java.util.Map
import java.util.Set
import org.eclipse.emf.common.util.EList
import org.eclipse.emf.common.util.URI
import org.eclipse.emf.ecore.EObject
import org.eclipse.emf.ecore.resource.Resource
import org.eclipse.xtext.EcoreUtil2
import org.eclipse.xtext.mwe.ResourceDescriptionsProvider
import org.eclipse.xtext.resource.IContainer
import org.eclipse.xtext.resource.IEObjectDescription
import org.eclipse.xtext.resource.IResourceDescription
import org.eclipse.xtext.resource.IResourceDescriptions
import org.eclipse.xtext.validation.Check

/**
 * Custom validation rules. 
 * 
 * see http://www.eclipse.org/Xtext/documentation.html#validation
 */
class DiaSwarmValidator extends AbstractDiaSwarmValidator {
	@Inject
	private ResourceDescriptionsProvider resourceDescriptionsProvider;

	@Inject
	private IContainer.Manager containerManager;

	@Check
	def checkStructureDeclarationStartsWithUpperCase(SourceType sourceType) {
		if (!Character::isUpperCase(sourceType.name.charAt(0)))
			warning("The name of a data type should start with a capital letter!",
				DiaswarmPackage.Literals::ABSTRACT_ELEMENT__NAME);
	}

	@Check
	def checkVariableDeclarationStartsWithLowerCase(DeviceType device) {
		if (!Character::isUpperCase(device.name.charAt(0)))
			warning("The name of a device should start with a capital letter!",
				DiaswarmPackage.Literals::ABSTRACT_ELEMENT__NAME);
	}

	@Check
	def checkContextDeclarationStartsWithUpperCase(ContextDef context) {
		if (!Character::isUpperCase(context.name.charAt(0)))
			warning("The name of a context should start with a capital letter!",
				DiaswarmPackage.Literals::ABSTRACT_ELEMENT__NAME);
	}

	@Check
	def checkControllerDeclarationStartsWithUpperCase(ControllerDef controller) {
		if (!Character::isUpperCase(controller.name.charAt(0)))
			warning("The name of a controller should start with a capital letter!",
				DiaswarmPackage.Literals::ABSTRACT_ELEMENT__NAME);
	}

	@Check
	def checkActionImplUniqueness(ActionImpl actionImpl) {
		val DeviceType deviceDef = (actionImpl.eContainer() as DeviceType);

		checkActionImplUniqueness(actionImpl, deviceDef, false);
	}

	def void checkActionImplUniqueness(ActionImpl actionImpl, DeviceType deviceDef, Boolean parent) {
		if (deviceDef != null) {
			var List<String> actionNames = new ArrayList<String>();
			var EList<ActionImpl> actions = deviceDef.getActions();

			for (ActionImpl actionImpl2 : actions) {
				if (!actionImpl.equals(actionImpl2)) {
					actionNames.add(actionImpl2.getAction().getName());
				}
			}

			if (actionNames.contains(actionImpl.getAction().getName())) {
				var String errorMessage = "";
				if (!parent) {
					errorMessage = "Duplicate action implementation.";
				} else {
					errorMessage = "Duplicate action implementation with the parent device '" + deviceDef.getName() +
						"'.";
				}
				error(errorMessage, DiaswarmPackage.Literals::ACTION_IMPL__ACTION);
			}

			checkActionImplUniqueness(actionImpl, deviceDef.superDevice, true);
		}
	}

	/**
	 * Check that the given method name is unique in the action
	 * 
	 * @param order
	 *            The order to check for uniqueness
	 */
	@Check
	def checkActionMethodUniqueness(OrderDef order) {
		val ActionDef parent = (Utils.getParentOfClass(order, ActionDef) as ActionDef);

		if (parent != null) {
			for (OrderDef order2 : parent.getOrders()) {
				if (order2.getName().equals(order.getName()) && order2 != order) {
					error("Duplicate method name.", DiaswarmPackage.Literals::ORDER_DEF__NAME);
				}
			}
		}
	}

	@Check
	def checkActionRefCorrectness(ActionRef actionRef) {
		for (ActionDef act : actionRef.device.allActions) {
			if (act == actionRef.action)
				return;
		}
		error("Action " + actionRef.action.getName() + " is not exposed by device " + actionRef.device.getName() + ".",
			DiaswarmPackage.Literals::ACTION_REF__ACTION);
	}

	def static List<ActionDef> getAllActions(DeviceType device) {
		var Set<ActionDef> processedActions = new HashSet<ActionDef>()

		if (device.superDevice != null)
			processedActions.addAll(getAllActions(device.superDevice))

		for (ActionImpl actionImpl : device.actions)
			processedActions.add(actionImpl.action)

		processedActions.toList
	}

	@Check
	def checkBehaviorActionRefUniqueness(ActionRef bcActionRef) {
		val ControllerBehaviorDef behavior = Utils.getBehaviorActionListContainer(bcActionRef);
		if (behavior != null && behavior.getActions().contains(bcActionRef)) {
			val String deviceName = bcActionRef.getDevice().getName();
			val String actionName = bcActionRef.getAction().getName();

			val EList<ActionRef> actionList = behavior.getActions();
			for (ActionRef bcActionRef2 : actionList) {
				if (!bcActionRef.equals(bcActionRef2)) {
					val String deviceName2 = bcActionRef2.getDevice().getName();
					val String actionName2 = bcActionRef2.getAction().getName();
					if (deviceName.equals(deviceName2) && actionName.equals(actionName2)) {
						error("Duplicate device action reference.", DiaswarmPackage.Literals::ACTION_REF__ACTION);
					}
				}
			}
		}
	}

	// Uniqueness of source refs in activation conditions through contracts in a
	// context (or controller)
	def static boolean providedEquals(DataRequirement source1, DataRequirement source2) {
		if (source1 == null) {
			return source2 == null;
		} else if (source2 == null) {
			return false;
		}

		if (source1.dataSource.context != null) {
			return source2.dataSource.context == source1.dataSource.context;
		} else if (source2.dataSource.context != null) {
			return false;
		}
		return source1.dataSource.sourceRef.device == source2.dataSource.sourceRef.device &&
			source1.dataSource.sourceRef.source == source2.dataSource.sourceRef.source;
	}

	@Check
	def checkContextBehaviorActivationUniqueness(InteractionContractDef behavior) {
		val ContextDef ctxt = Utils.getContextContainer(behavior);
		for (InteractionContractDef bDef : ctxt.getBehaviors()) {
			if (bDef != behavior) {
				if (behavior.interaction.equals("required")) {
					if (bDef.interaction.equals("required")) {
						error("Duplicate activation condition in context contract.",
							DiaswarmPackage.Literals::INTERACTION_CONTRACT_DEF__CONTRACT);
					}
				} else {
					if (providedEquals((bDef.contract as PushInteractionContract).providedData,
						(behavior.contract as PushInteractionContract).providedData)) {
						error("Duplicate activation condition in context contract.",
							DiaswarmPackage.Literals::INTERACTION_CONTRACT_DEF__CONTRACT);
					}
				}
			}
		}
	}

	@Check
	def checkControllerBehaviorActivationUniqueness(ControllerBehaviorDef behavior) {
		val ControllerDef ctrl = Utils.getControllerContainer(behavior);
		for (ControllerBehaviorDef bDef : ctrl.getBehaviors()) {
			if (bDef != behavior) {
				if (bDef.getActivationCondition() == behavior.getActivationCondition()) {
					error("Duplicate activation condition in controller contract.",
						DiaswarmPackage.Literals::CONTROLLER_BEHAVIOR_DEF__ACTIVATION_CONDITION);
				}
			}
		}

	}

	@Check
	def checkStructFieldDefUniqueness(StructFieldDef structFieldDef) {
		val String structFieldName = structFieldDef.getName();
		val StructDef structDef = Utils.getStructDefContainer(structFieldDef);
		val EList<StructFieldDef> structFields = structDef.getStructFields();

		for (StructFieldDef structFieldDef2 : structFields) {
			if (!structFieldDef.equals(structFieldDef2)) {
				val String structFieldName2 = structFieldDef2.getName();
				if (structFieldName.equals(structFieldName2)) {
					error("Duplicate field name.", DiaswarmPackage.Literals::STRUCT_FIELD_DEF__NAME);
				}
			}
		}
	}

	def List<URI> getImportedURI(Resource parent) {
		val DomainModel doc = Utils.getDocumentContainer(parent.getContents().get(0));
		var List<URI> result = new LinkedList<URI>();
		result.add(parent.getURI());
		for (IncludeSpec incl : doc.getIncludes()) {
			val URI importUri = URI.createURI(incl.getImportURI());
			if (EcoreUtil2.isValidUri(parent, importUri)) {
				result.add(importUri);
			}
		}
		return result;
	}

	def boolean checkObjectNameUniqueness(EObject object, String name, String... compatibleTypes) {
		val Resource resource = object.eResource();
		val List<URI> importedURIs = getImportedURI(resource);
		val IResourceDescriptions resourceDescriptions = resourceDescriptionsProvider.get(resource.getResourceSet());
		val IResourceDescription resourceDescription = resourceDescriptions.getResourceDescription(resource.getURI());

		for (IContainer c : containerManager.getVisibleContainers(resourceDescription, resourceDescriptions)) {
			for (IEObjectDescription desc : c.getExportedObjects()) {
				if (importedURIs.contains(desc.getEObjectOrProxy().eResource().getURI()) &&
					objectNameMatchesDescription(object, name, desc, compatibleTypes))
					return false;
			}
		}
		return true;
	}

	def boolean objectNameMatchesDescription(EObject object, String name, IEObjectDescription desc,
		String... compatibleTypes) {
		if (desc.getQualifiedName().toString().equals(name) && !desc.getEObjectOrProxy().equals(object)) {
			// Check for type
			if (desc.getEClass().equals(object.eClass()))
				return true;
			for (String s : compatibleTypes) {
				if (desc.getEClass().getName().equals(s))
					return true;
			}
		}
		return false;
	}

	/**
	 * Check for duplicate action name
	 * 
	 * @param action
	 *            The action to do check on
	 */
	@Check
	def void checkDuplicateActionName(ActionDef action) {
		if (!checkObjectNameUniqueness(action, action.getName())) {
			error("Duplicate action name.", DiaswarmPackage.Literals::ACTION_DEF__NAME);
		}
	}

	/**
	 * Check for duplicate device name
	 * 
	 * @param device
	 *            The device to do check on
	 */
	@Check
	def void checkDuplicateDeviceName(DeviceType device) {
		if (!checkObjectNameUniqueness(device, device.getName(), "ContextDef", "ControllerDef")) {
			error("Duplicate device name.", DiaswarmPackage.Literals::DEVICE_TYPE__SUPER_DEVICE);
		}
	}

	/**
	 * Check for duplicate controller name
	 * 
	 * @param controller
	 *            The controller to do check on
	 */
	@Check
	def checkDuplicateControllerName(ControllerDef controller) {
		if (!checkObjectNameUniqueness(controller, controller.getName(), "DeviceDef", "ContextDef")) {
			error("Duplicate controller name.", DiaswarmPackage.Literals::CONTROLLER_DEF__BEHAVIORS);
		}
	}

	/**
	 * Check for duplicate context name
	 * 
	 * @param context
	 *            The context to do check on
	 */
	@Check
	def checkDuplicateContextName(ContextDef context) {
		if (!checkObjectNameUniqueness(context, context.getName(), "DeviceDef", "ControllerDef")) {
			error("Duplicate context name.", DiaswarmPackage.Literals::CONTEXT_DEF__DATA_TYPE_REF);
		}
	}

	def checkDuplicateSourceAttrInDevice(DeviceType device, String name, EObject obj) {
		if (device != null) {
			for (SourceDef sourceDef : device.getSources()) {
				if (sourceDef != obj && sourceDef.name.equals(name))
					return true;
			}
			for (AttributeDef attrDef : device.getAttributes()) {
				if (attrDef != obj && attrDef.name.equals(name))
					return true;
			}
		}
		return false;
	}

	/**
	 * Check for duplicate source/attribute in a device
	 */
	@Check
	def checkDuplicateSourceInDevice(SourceDef sourceDef) {
		val DeviceType device = (Utils.getParentOfClass(sourceDef, DeviceType) as DeviceType);
		if (device != null) {
			if (checkDuplicateSourceAttrInDevice(device, sourceDef.getName(), sourceDef))
				error("Duplicate source name.", DiaswarmPackage.Literals::SOURCE_DEF__NAME);
			// Check for super devices
			val Set<DeviceType> superDevices = Utils.getAllSuperDevices(device);
			for (DeviceType superDevice : superDevices) {
				if (checkDuplicateSourceAttrInDevice(superDevice, sourceDef.getName(), sourceDef))
					error(
						"Source name is duplicated with a source or an attribute from the super device " +
							superDevice.getName() + ".", DiaswarmPackage.Literals::SOURCE_DEF__NAME);
			}
		}
	}

	/**
	 * Check for duplicate source/attribute in a device
	 */
	@Check
	def checkDuplicateAttributeInDevice(AttributeDef attributeDef) {
		val DeviceType device = (Utils.getParentOfClass(attributeDef, DeviceType) as DeviceType);
		if (device != null) {
			if (checkDuplicateSourceAttrInDevice(device, attributeDef.getName(), attributeDef))
				error("Duplicate attribute name.", DiaswarmPackage.Literals::ATTRIBUTE_DEF__NAME);
			// Check for super device
			val Set<DeviceType> superDevices = Utils.getAllSuperDevices(device);
			for (DeviceType superDevice : superDevices) {
				if (checkDuplicateSourceAttrInDevice(superDevice, attributeDef.getName(), attributeDef))
					error(
						"Attribute name is duplicated with a source or an attribute from the super device " +
							superDevice.getName() + ".", DiaswarmPackage.Literals::ATTRIBUTE_DEF__NAME);
			}
		}
	}

	/**
	 * Check for duplicate type name
	 * 
	 * @param type
	 *            The type to do check on
	 */
	@Check
	def checkDuplicateTypeNameStructure(SourceType type) {
		if (!checkObjectNameUniqueness(type, type.getName(), "StructDef", "EnumDef", "ImportHostType")) {
			error("Duplicate type name.", DiaswarmPackage.Literals::STRUCT_DEF__STRUCT_FIELDS);
		}
	}

	/**
	 * Self-Extends test
	 */
	@Check
	def checkSelfExtends(DeviceType device) {
		val Set<DeviceType> superDevices = Utils.getAllSuperDevices(device);
		if (superDevices.contains(device)) {
			error("There is a cycle in device inheritance.", DiaswarmPackage.Literals::DEVICE_TYPE__SUPER_DEVICE);
			return;
		}
	}
	
	@Check
	def checkIdInDevices(DeviceType device) {
		val id = device.allAttributes.filter[name.equals("id") && attrTypeRef.primitiveTypeRef.literal.equals("String") ].size
		if(id == 0)
			error("A device has to define an id attribute of type String !", DiaswarmPackage.Literals::DEVICE_TYPE__ATTRIBUTES);
	}
	
	def static List<AttributeDef> getAllAttributes(DeviceType device) {
		var Set<AttributeDef> processedActions = new HashSet<AttributeDef>()

		if (device.superDevice != null)
			processedActions.addAll(getAllAttributes(device.superDevice))

		for (AttributeDef attribute : device.attributes)
			processedActions.add(attribute)

		processedActions.toList
	}
	
	@Check
	def checkOneMRJobPerContext(ContextDef context) {
		var mrJobCount = 0;
		
		for(InteractionContractDef interactionContract : context.behaviors) {
			if(interactionContract.interaction.equals("periodic"))
				if((interactionContract.contract as PushInteractionContract).providedData.map)
					mrJobCount++;
		}
		
		if(mrJobCount > 1)
		error("Multiple MapReduce jobs are not allowed in a context.", DiaswarmPackage.Literals::CONTEXT_DEF__BEHAVIORS);
	}
	
	@Check
	def checkGroupByAttributeWhenMR(ContextDef context) {
		
		for(InteractionContractDef interactionContract : context.behaviors) {
			if(interactionContract.interaction.equals("periodic"))
				if((interactionContract.contract as PushInteractionContract).providedData.map)
					if((interactionContract.contract as PushInteractionContract).providedData.groupingAttribute == null)
					error("Map reduce requires a grouping attribute.", DiaswarmPackage.Literals::CONTEXT_DEF__BEHAVIORS);
		}
	}
	
	@Check
	def checkPeriodicityWhenPeriodicInteraction(ContextDef context) {
		for(InteractionContractDef interactionContract : context.behaviors) {
			if(interactionContract.interaction.equals("periodic"))
				if((interactionContract.contract as PushInteractionContract).providedData.dataSource.sourceRef != null)
					if((interactionContract.contract as PushInteractionContract).providedData.dataSource.sourceRef.periodicity == null)
					error("Data delivery frequency is not defined.", DiaswarmPackage.Literals::CONTEXT_DEF__BEHAVIORS);
		}
	}
	
	
}
